/*
 * @version 2017/03/03
 * - print warning about scaling large images
 * @version 2017/03/02
 * - improved nullness checking
 * - bug fixes for scale() method
 * @version 2016/05/21
 * - initial version
 */

package stanford.androidlib;

import android.content.*;
import android.graphics.*;
import android.support.annotation.*;
import android.util.Log;
import android.view.View;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;

/**
 * A utility class for loading and transforming Bitmap objects.
 *
 * <pre>
 * Bitmap bitmap1 = SimpleBitmap.with(context).get(R.drawable.foo);
 * Bitmap bitmap2 = SimpleBitmap.with(context).scale(R.drawable.foo, getWidth()/2, getHeight()/2);
 * </pre>
 */
public class SimpleBitmap {
    private static Context context = null;
    private static final SimpleBitmap INSTANCE = new SimpleBitmap();
    private static int MAX_BITMAP_WIDTH = -1;   // max size of a bitmap
    private static int MAX_BITMAP_HEIGHT = -1;
    private static int MAX_TEXTURE_SIZE = 2048;   // Android internal limit for many bitmaps
    private boolean filter = true;

    /**
     * Returns a singleton SimpleBitmap instance bound to the given context.
     */
    public static SimpleBitmap with(@NonNull Context context) {
        SimpleBitmap.context = context;
        checkMaxBitmapSize();
        return INSTANCE;
    }

    /**
     * Returns a singleton SimpleBitmap instance bound to the given view's context.
     */
    public static SimpleBitmap with(@NonNull View context) {
        Context newContext = context.getContext();
        if (SimpleBitmap.context == null || newContext != null) {
            SimpleBitmap.context = newContext;
        }
        checkMaxBitmapSize();
        return INSTANCE;
    }

    private SimpleBitmap() {
        // empty
    }

    /*
     * Helper to find out maximum allowable bitmap size
     */
    private static void checkMaxBitmapSize() {
        // test-and-test-and-set
        if (MAX_BITMAP_HEIGHT < 0 || MAX_BITMAP_WIDTH < 0) {
            synchronized (SimpleBitmap.class) {
                if (MAX_BITMAP_HEIGHT < 0 || MAX_BITMAP_WIDTH < 0) {
                    Canvas canvas = new Canvas();
                    MAX_BITMAP_WIDTH = canvas.getMaximumBitmapWidth();
                    MAX_BITMAP_HEIGHT = canvas.getMaximumBitmapHeight();
                }
            }
        }
    }

    /**
     * Returns the bitmap image for the resource file with the given ID.
     * @throws NullPointerException if the image is missing or cannot be loaded.
     */
    public Bitmap get(@DrawableRes int id) {
        Bitmap bitmap = BitmapFactory.decodeResource(context.getResources(), id);
        if (bitmap == null) {
            throw new NullPointerException("Image not found or could not be decoded: ID " + id);
        }
        return bitmap;
    }

    /**
     * Returns the bitmap image for the file located at the given web URL.
     * @throws IORuntimeException if the URL cannot be read or is not a valid image.
     * @throws NullPointerException if the image is missing or cannot be loaded.
     */
    public Bitmap get(@NonNull String url) {
        try {
            URL theUrl = new URL(url);
            Bitmap bitmap = BitmapFactory.decodeStream(theUrl.openStream());
            if (bitmap == null) {
                throw new NullPointerException("Image not found or could not be decoded: URL=" + url);
            }
            return bitmap;
        } catch (MalformedURLException mfurle) {
            throw new IllegalArgumentException("Invalid URL: " + url, mfurle);
        } catch (IOException ioe) {
            throw new IORuntimeException(ioe);
        }
    }

    /**
     * Returns the bitmap image for the file located at the given web URL.
     * @throws IORuntimeException if the URL cannot be read or is not a valid image.
     * @throws NullPointerException if the image is missing or cannot be loaded.
     */
    public Bitmap get(@NonNull URL url) {
        try {
            Bitmap bitmap = BitmapFactory.decodeStream(url.openStream());
            if (bitmap == null) {
                throw new NullPointerException("Image not found or could not be decoded: URL=" + url);
            }
            return bitmap;
        } catch (IOException ioe) {
            throw new IORuntimeException(ioe);
        }
    }

    /**
     * Returns the bitmap that corresponds to the given resource ID,
     * scaled to the given width and height.
     * @throws NullPointerException if the image is missing or cannot be loaded.
     */
    public Bitmap get(@DrawableRes int id, float width, float height) {
        return scale(id, width, height);
    }

    /**
     * Returns a list of the bitmap images for the resource files with the given IDs.
     * @throws NullPointerException if any image is missing or cannot be loaded.
     */
    public ArrayList<Bitmap> getAll(@DrawableRes int... ids) {
        ArrayList<Bitmap> list = new ArrayList<>();
        for (@DrawableRes int id : ids) {
            list.add(get(id));
        }
        return list;
    }

    /**
     * Returns a list of the bitmap images for the resource files with the given IDs,
     * scaled by the given factor.
     * @throws NullPointerException if any image is missing or cannot be loaded.
     */
    public ArrayList<Bitmap> getAllScaled(float scaleFactor, @DrawableRes int... ids) {
        ArrayList<Bitmap> list = new ArrayList<>();
        for (@DrawableRes int id : ids) {
            list.add(scale(id, scaleFactor));
        }
        return list;
    }

    /**
     * Returns a list of the bitmap images for the resource files with the given IDs,
     * scaled to the given width and height.
     * @throws NullPointerException if any image is missing or cannot be loaded.
     */
    public ArrayList<Bitmap> getAllScaled(float width, float height, @DrawableRes int... ids) {
        ArrayList<Bitmap> list = new ArrayList<>();
        for (@DrawableRes int id : ids) {
            list.add(scale(id, width, height));
        }
        return list;
    }

    /**
     * Whether we filter bitmaps on scale/rotate for smoothness.
     * Slower but better-looking. On by default.
     */
    public boolean isFiltered() {
        return filter;
    }

    /**
     * Returns a new bitmap based on the given bitmap but rotated by the given number of degrees
     * clockwise about its center point.
     */
    public Bitmap rotate(@NonNull Bitmap bitmap, float degrees) {
        float centerX = bitmap.getWidth() / 2f;
        float centerY = bitmap.getHeight() / 2f;
        return rotate(bitmap, degrees, centerX, centerY);
    }

    /**
     * Returns a new bitmap based on the given bitmap but rotated by the given number of degrees
     * clockwise about its center point.
     */
    public Bitmap rotate(@DrawableRes int id, float degrees) {
        Bitmap bitmap = get(id);
        return rotate(bitmap, degrees);
    }

    /**
     * Returns a new bitmap based on the given bitmap but rotated by the given number of degrees
     * clockwise about the point (rx, ry).
     */
    public Bitmap rotate(@NonNull Bitmap bitmap, float degrees, float rx, float ry) {
        Matrix matrix = new Matrix();
        matrix.postRotate(degrees, rx, ry);
        return Bitmap.createBitmap(bitmap,
                0, 0, bitmap.getWidth(), bitmap.getHeight(),
                matrix, /* filter */ true);
    }

    /**
     * Returns a new bitmap based on the given bitmap but rotated by the given number of degrees
     * clockwise about the point (rx, ry).
     */
    public Bitmap rotate(@DrawableRes int id, float degrees, float rx, float ry) {
        return rotate(get(id), degrees, rx, ry);
    }

    /**
     * Returns a new bitmap based on the given bitmap but rotated counter-clockwise by 90 degrees
     * about its center point.
     */
    public Bitmap rotateLeft(@NonNull Bitmap bitmap) {
        return rotate(bitmap, /* degrees */ -90);
    }

    /**
     * Returns a new bitmap based on the given bitmap but rotated counter-clockwise by 90 degrees
     * about its center point.
     */
    public Bitmap rotateLeft(@DrawableRes int id) {
        Bitmap bitmap = get(id);
        return rotate(bitmap, /* degrees */ -90);
    }

    /**
     * Returns a new bitmap based on the given bitmap but rotated clockwise by 90 degrees
     * about its center point.
     */
    public Bitmap rotateRight(@NonNull Bitmap bitmap) {
        return rotate(bitmap, /* degrees */ 90);
    }

    /**
     * Returns a new bitmap based on the given bitmap but rotated clockwise by 90 degrees
     * about its center point.
     */
    public Bitmap rotateRight(@DrawableRes int id) {
        Bitmap bitmap = get(id);
        return rotate(bitmap, /* degrees */ 90);
    }

    /**
     * Returns a new bitmap which is the given bitmap resized to the given size.
     * @throws NullPointerException if the image is missing or cannot be scaled.
     */
    public Bitmap scale(@NonNull Bitmap bitmap, float width, float height) {
        if (width > MAX_BITMAP_WIDTH || height > MAX_BITMAP_HEIGHT) {
            throw new IllegalArgumentException("cannot scale to " + (int) width + "x" + (int) height
                    + "; max allowed size is " + MAX_BITMAP_WIDTH + "x" + MAX_BITMAP_HEIGHT);
        } else if (width > MAX_TEXTURE_SIZE || height > MAX_TEXTURE_SIZE) {
            // print warning about large images
            Log.w("SimpleBitmap", "WARNING: You are trying to scale an image to size "
                    + (int) width + "x" + (int) height
                    + "; some devices will fail to generate any image larger than "
                    + MAX_TEXTURE_SIZE + "x" + MAX_TEXTURE_SIZE);
        }
        Bitmap scaled = Bitmap.createScaledBitmap(bitmap, (int) width, (int) height, /* filter */ isFiltered());
        if (scaled == null) {
            throw new NullPointerException("Image not found or could not be scaled");
        }
        return scaled;
    }

    /**
     * Returns the bitmap that corresponds to the given resource ID.
     */
    public Bitmap scale(@DrawableRes int id, float width, float height) {
        return scale(get(id), width, height);
    }

    /**
     * Returns a new bitmap which is the given bitmap resized by the given
     * factor.  For example, if the scale factor is 0.5f, the image will
     * shrink to half its current size.
     */
    public Bitmap scale(@NonNull Bitmap bitmap, float scaleFactor) {
        float width = bitmap.getWidth() * scaleFactor;
        float height = bitmap.getHeight() * scaleFactor;
        return scale(bitmap, width, height);
    }

    /**
     * Returns a new bitmap which is the given bitmap resized by the given
     * factor.  For example, if the scale factor is 0.5f, the image will
     * shrink to half its current size.
     */
    public Bitmap scale(@DrawableRes int id, float scaleFactor) {
        Bitmap bitmap = get(id);
        return scale(bitmap, scaleFactor);
    }

    /**
     * Returns a new bitmap which is the given bitmap resized to the largest size
     * that can fit within the given width/height while maintaining its aspect ratio.
     */
    public Bitmap scaleToFit(@NonNull Bitmap bitmap, float width, float height) {
        // example: image is  40x 30, fit to 200x400
        //                => 200x150
        // example: image is 400x300, fit to 200x400
        //                => 200x150
        // example: image is 400x300, fit to 1200x450
        //                => 600x450
        float widthRatio  = bitmap.getWidth() / width;
        float heightRatio = bitmap.getHeight() / height;
        if (widthRatio > heightRatio) {
            float scaleFactor = width / bitmap.getWidth();
            return scale(bitmap, scaleFactor);
        } else {
            float scaleFactor = height / bitmap.getHeight();
            return scale(bitmap, scaleFactor);
        }
    }

    /**
     * Returns a new bitmap which is the given bitmap resized to the largest size
     * that can fit within the given width/height while maintaining its aspect ratio.
     */
    public Bitmap scaleToFit(@DrawableRes int id, float width, float height) {
        return scaleToFit(get(id), width, height);
    }

    /**
     * Returns a new bitmap which is the given bitmap resized to the largest size
     * that can fit within the given view's width/height while maintaining its aspect ratio.
     */
    public Bitmap scaleToFit(@NonNull Bitmap bitmap, @NonNull View view) {
        return scaleToFit(bitmap, view.getWidth(), view.getHeight());
    }

    /**
     * Returns a new bitmap which is the given bitmap resized to the largest size
     * that can fit within the given view's width/height while maintaining its aspect ratio.
     */
    public Bitmap scaleToFit(@DrawableRes int id, @NonNull View view) {
        return scaleToFit(get(id), view);
    }

    /**
     * Returns a new bitmap which is the given bitmap resized to the largest size
     * that can fit within the given canvas's width/height while maintaining its aspect ratio.
     */
    public Bitmap scaleToFit(@NonNull Bitmap bitmap, @NonNull Canvas canvas) {
        return scaleToFit(bitmap, canvas.getWidth(), canvas.getHeight());
    }

    /**
     * Returns a new bitmap which is the given bitmap resized to the largest size
     * that can fit within the given canvas's width/height while maintaining its aspect ratio.
     */
    public Bitmap scaleToFit(@DrawableRes int id, @NonNull Canvas canvas) {
        return scaleToFit(get(id), canvas);
    }

    /**
     * Returns a new bitmap which is the given bitmap resized to have
     * the given height, and a proportionally scaled width to match.
     */
    public Bitmap scaleToHeight(@NonNull Bitmap bitmap, float height) {
        float width = bitmap.getWidth() * height / bitmap.getHeight();
        return scale(bitmap, width, height);
    }

    /**
     * Returns the bitmap that corresponds to the given resource ID.
     */
    public Bitmap scaleToHeight(@DrawableRes int id, float height) {
        return scaleToHeight(get(id), height);
    }

    /**
     * Returns a new bitmap which is the given bitmap resized to have
     * the given width, and a proportionally scaled height to match.
     */
    public Bitmap scaleToWidth(@NonNull Bitmap bitmap, float width) {
        float height = bitmap.getHeight() * width / bitmap.getWidth();
        return scale(bitmap, width, height);
    }

    /**
     * Returns the bitmap that corresponds to the given resource ID.
     */
    public Bitmap scaleToWidth(@DrawableRes int id, float width) {
        return scaleToWidth(get(id), width);
    }

    /**
     * Whether we filter bitmaps on scale/rotate for smoothness.
     * Slower but better-looking. On by default.
     */
    public SimpleBitmap setFiltered(boolean filter) {
        this.filter = filter;
        return this;
    }
}
