/*
 * @version 2017/03/03
 * - added format(), get(), pluralize()
 * - added Locale version of many methods
 * - renamed some methods, e.g. localizeCurrency() -> currency()
 * @version 2016/03/02
 * - initial version
 */

package stanford.androidlib;

import android.content.Context;
import android.content.res.Configuration;
import android.os.Build;
import android.support.annotation.*;
import android.view.View;
import java.text.*;
import java.util.*;

/**
 * Utility class to help with localization of common values.
 */
public final class SimpleLocalization {
    private static Context context;
    private static SimpleLocalization INSTANCE = new SimpleLocalization();

    /*
     * You cannot directly construct an object of type SimpleLocalization.
     */
    private SimpleLocalization() {
        // empty
    }

    /*
     * Returns an object of type SimpleLocalization that is bound to the given context.
     * Pass your activity, fragment, etc. as the context.
     */
    public static SimpleLocalization with(Context context) {
        SimpleLocalization.context = context;
        return INSTANCE;
    }

    /**
     * Returns a singleton SimpleLocalization instance bound to the given view's context.
     */
    public static SimpleLocalization with(View context) {
        Context newContext = context.getContext();
        if (SimpleLocalization.context == null || newContext != null) {
            SimpleLocalization.context = newContext;
        }
        return INSTANCE;
    }

    /**
     * Returns a currency number string formatted for this locale.
     */
    public String currency(double amount) {
        return currency(Currency.getInstance(Locale.getDefault()), amount);
    }

    /**
     * Returns a currency number string formatted for the given currency.
     */
    public String currency(String currency, double amount) {
        return currency(Currency.getInstance(currency), amount);
    }

    /**
     * Returns a currency number string formatted for the given currency.
     */
    public String currency(Currency currency, double amount) {
        NumberFormat money = NumberFormat.getCurrencyInstance();
        money.setCurrency(Currency.getInstance("EUR"));
        return money.format(amount);
    }

    /**
     * Returns a currency number string formatted in the currency of the given locale.
     */
    public String currency(double amount, Locale locale) {
        return currency(Currency.getInstance(locale), amount);
    }

    /**
     * Returns a date string formatted for this locale.
     */
    public String date(Date date) {
        return date(date, Locale.getDefault());
    }

    /**
     * Returns a date string formatted for this locale.
     */
    public String date(Date date, Locale locale) {
        DateFormat fmt = DateFormat.getDateTimeInstance(DateFormat.DEFAULT, DateFormat.DEFAULT, locale);
        return fmt.format(date);
    }

    /**
     * Formats a resource string for the given ID and fills it in with the given placeholders.
     */
    public String format(@StringRes int id, Object... args) {
        return context.getResources().getString(id, args);
    }

    /**
     * Returns a resource string for the given ID.
     */
    public String get(@StringRes int id) {
        return context.getResources().getString(id);
    }

    /**
     * Formats a resource string for the given ID and fills it in with the given placeholders.
     */
    public String get(@StringRes int id, Object... args) {
        return context.getResources().getString(id, args);
    }

    /**
     * Returns true if this app is running in a Left-to-Right (LTR) layout.
     */
    public boolean isLTR() {
        return !isRTL();
    }

    /**
     * Returns true if the given locale reads Left-to-Right (LTR).
     */
    public boolean isLTR(Locale locale) {
        return !isRTL(locale);
    }

    /**
     * Returns true if this app is running in a Right-to-Left (RTL) layout.
     */
    public boolean isRTL() {
        if (context != null && Build.VERSION.SDK_INT >= Build.VERSION_CODES.JELLY_BEAN_MR1) {
            Configuration config = context.getResources().getConfiguration();
            return config.getLayoutDirection() == View.LAYOUT_DIRECTION_RTL;
        } else {
            final int directionality = Character.getDirectionality(Locale.getDefault().getDisplayName().charAt(0));
            return directionality == Character.DIRECTIONALITY_RIGHT_TO_LEFT ||
                    directionality == Character.DIRECTIONALITY_RIGHT_TO_LEFT_ARABIC;
        }
    }

    /**
     * Returns true if the given locale reads Right-to-Left (RTL).
     */
    public boolean isRTL(Locale locale) {
        final int directionality = Character.getDirectionality(locale.getDisplayName().charAt(0));
        return directionality == Character.DIRECTIONALITY_RIGHT_TO_LEFT ||
                directionality == Character.DIRECTIONALITY_RIGHT_TO_LEFT_ARABIC;
    }

    /**
     * Returns a number string formatted for this locale.
     */
    public String number(double number) {
        NumberFormat fmt = NumberFormat.getInstance();
        return fmt.format(number);
    }

    /**
     * Returns a number string formatted for the given locale.
     */
    public String number(double number, Locale locale) {
        NumberFormat fmt = NumberFormat.getInstance(locale);
        return fmt.format(number);
    }

    /**
     * Returns a number string formatted for this locale.
     */
    public String number(int number) {
        NumberFormat fmt = NumberFormat.getInstance();
        return fmt.format(number);
    }

    /**
     * Returns a number string formatted for the given locale.
     */
    public String number(int number, Locale locale) {
        NumberFormat fmt = NumberFormat.getInstance(locale);
        return fmt.format(number);
    }

    /**
     * Returns a number string formatted for this locale.
     */
    public String number(long number) {
        NumberFormat fmt = NumberFormat.getInstance();
        return fmt.format(number);
    }

    /**
     * Returns a number string formatted for the given locale.
     */
    public String number(long number, Locale locale) {
        NumberFormat fmt = NumberFormat.getInstance(locale);
        return fmt.format(number);
    }

    /**
     * Treats the given text as a real number in the current locale,
     * parses it into a double, and returns it.
     */
    public double parseLocalizedDouble(String text) {
        return parseLocalizedDouble(text, Locale.getDefault());
    }

    /**
     * Treats the given text as a real number in the given locale,
     * parses it into a double, and returns it.
     */
    public double parseLocalizedDouble(String text, Locale locale) {
        NumberFormat fmt = NumberFormat.getInstance(locale);
        try {
            return (Double) fmt.parse(text);
        } catch (ClassCastException | ParseException e) {
            throw new IllegalArgumentException("Unable to parse as localized double: \"" + text + "\"");
        }
    }

    /**
     * Treats the given text as a real number in the current locale,
     * parses it into a float, and returns it.
     */
    public float parseLocalizedFloat(String text) {
        return parseLocalizedFloat(text, Locale.getDefault());
    }

    /**
     * Treats the given text as a real number in the given locale,
     * parses it into a float, and returns it.
     */
    public float parseLocalizedFloat(String text, Locale locale) {
        NumberFormat fmt = NumberFormat.getInstance(locale);
        try {
            return (Float) fmt.parse(text);
        } catch (ClassCastException | ParseException e) {
            throw new IllegalArgumentException("Unable to parse as localized float: \"" + text + "\"");
        }
    }

    /**
     * Treats the given text as an integer in the current locale,
     * parses it into an int, and returns it.
     */
    public int parseLocalizedInt(String text) {
        return parseLocalizedInt(text, Locale.getDefault());
    }

    /**
     * Treats the given text as an integer in the given locale,
     * parses it into an int, and returns it.
     */
    public int parseLocalizedInt(String text, Locale locale) {
        NumberFormat fmt = NumberFormat.getInstance(locale);
        try {
            return (Integer) fmt.parse(text);
        } catch (ClassCastException | ParseException e) {
            throw new IllegalArgumentException("Unable to parse as localized int: \"" + text + "\"");
        }
    }

    /**
     * Treats the given text as a long integer in the current locale,
     * parses it into a long, and returns it.
     */
    public long parseLocalizedLong(String text) {
        return parseLocalizedLong(text, Locale.getDefault());
    }

    /**
     * Treats the given text as a long integer in the given locale,
     * parses it into a long, and returns it.
     */
    public long parseLocalizedLong(String text, Locale locale) {
        NumberFormat fmt = NumberFormat.getInstance(locale);
        try {
            return (Long) fmt.parse(text);
        } catch (ClassCastException | ParseException e) {
            throw new IllegalArgumentException("Unable to parse as localized long: \"" + text + "\"");
        }
    }

    /**
     * Formats a "quantity" resource string for the given ID, using the right pluralization for the
     * given quantity, and fills it in with the given placeholders.
     */
    public String pluralize(@PluralsRes int id, int quantity, Object... args) {
        return context.getResources().getQuantityString(id, quantity, args);
    }
}
